#include <iostream>
#include <cstdlib>
#include <ctime>
#include <cmath>
#include <cstring>
#include <immintrin.h>

#ifdef SINGLE
#define VECTOR_SIZE 16
#else
#define VECTOR_SIZE 8
#endif

#define MATRIX_SIZE (VECTOR_SIZE*VECTOR_SIZE)

#ifdef SINGLE
typedef float REAL;
typedef __m512 type512;
#else
typedef double REAL;
typedef __m512d type512;
#endif

#define DIMM 1000000

#ifdef SINGLE
#define DIM DIMM
#else
#define DIM DIMM*4
#endif

#define ALIGNMENT 64

REAL* matrixArray;
REAL* vectorArray;
REAL* vectorResultArray;

type512 zmm00;
type512 zmm01;
type512 zmm02;
type512 zmm03;
type512 zmm04;
type512 zmm05;
type512 zmm06;
type512 zmm07;
type512 zmm08;
type512 zmm09;
type512 zmm10;
type512 zmm11;
type512 zmm12;
type512 zmm13;
type512 zmm14;
type512 zmm15;
type512 zmm16;
type512 zmm17;
type512 zmm18;
type512 zmm19;
type512 zmm20;
type512 zmm21;
type512 zmm22;
type512 zmm23;
type512 zmm24;
type512 zmm25;
type512 zmm26;
type512 zmm27;
type512 zmm28;
type512 zmm29;
type512 zmm30;
type512 zmm31;

typedef void (* FUNC)(REAL *, REAL *, REAL *);

float MeasuredRun(REAL* imtx, REAL* ivec, REAL* ires, FUNC func)
{
	clock_t t = clock();
	for (int i = 0; i < DIM; i++)
	{
		func(imtx + i*MATRIX_SIZE, ivec + i*VECTOR_SIZE, ires + i*VECTOR_SIZE);
	}
	t = clock()-t;
	return ((float)t)/CLOCKS_PER_SEC;
}


void MVmDirect(REAL* mtx, REAL* vec, REAL* res)
{
	for (int i = 0; i < VECTOR_SIZE; i++)
	{
		res[i] = 0.;
		for (int j = 0; j < VECTOR_SIZE; j++)
		{
			res[i] += mtx[i + j*VECTOR_SIZE] * vec[j];
		}
	}
}

void MVmDirectInv(REAL* mtx, REAL* vec, REAL* res)
{
	memset(res, 0., sizeof(REAL)*VECTOR_SIZE);

	for (int j = 0; j < VECTOR_SIZE; j++)
	{
		for (int i = 0; i < VECTOR_SIZE; i++)
		{
			res[i] += mtx[i + j*VECTOR_SIZE] * vec[j];
		}
	}
}

void MVmSimdIntrinsic(REAL* mtx, REAL* vec, REAL* res)
{
#ifdef SINGLE
	// matrix
	zmm00 = _mm512_load_ps(mtx);
	zmm01 = _mm512_load_ps(mtx + VECTOR_SIZE);
	zmm02 = _mm512_load_ps(mtx + VECTOR_SIZE * 2);
	zmm03 = _mm512_load_ps(mtx + VECTOR_SIZE * 3);
	zmm04 = _mm512_load_ps(mtx + VECTOR_SIZE * 4);
	zmm05 = _mm512_load_ps(mtx + VECTOR_SIZE * 5);
	zmm06 = _mm512_load_ps(mtx + VECTOR_SIZE * 6);
	zmm07 = _mm512_load_ps(mtx + VECTOR_SIZE * 7);
	zmm08 = _mm512_load_ps(mtx + VECTOR_SIZE * 8);
	zmm09 = _mm512_load_ps(mtx + VECTOR_SIZE * 9);
	zmm10 = _mm512_load_ps(mtx + VECTOR_SIZE * 10);
	zmm11 = _mm512_load_ps(mtx + VECTOR_SIZE * 11);
	zmm12 = _mm512_load_ps(mtx + VECTOR_SIZE * 12);
	zmm13 = _mm512_load_ps(mtx + VECTOR_SIZE * 13);
	zmm14 = _mm512_load_ps(mtx + VECTOR_SIZE * 14);
	zmm15 = _mm512_load_ps(mtx + VECTOR_SIZE * 15);

	// vector
	zmm16 = _mm512_extload_ps(vec, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm17 = _mm512_extload_ps(vec + 1, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm18 = _mm512_extload_ps(vec + 2, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm19 = _mm512_extload_ps(vec + 3, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm20 = _mm512_extload_ps(vec + 4, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm21 = _mm512_extload_ps(vec + 5, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm22 = _mm512_extload_ps(vec + 6, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm23 = _mm512_extload_ps(vec + 7, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm24 = _mm512_extload_ps(vec + 8, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm25 = _mm512_extload_ps(vec + 9, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm26 = _mm512_extload_ps(vec + 10, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm27 = _mm512_extload_ps(vec + 11, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm28 = _mm512_extload_ps(vec + 12, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm29 = _mm512_extload_ps(vec + 13, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm30 = _mm512_extload_ps(vec + 14, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm31 = _mm512_extload_ps(vec + 15, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);

	zmm00 = _mm512_mul_ps(zmm00, zmm16);
	zmm00 = _mm512_fmadd_ps(zmm01, zmm17, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm02, zmm18, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm03, zmm19, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm04, zmm20, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm05, zmm21, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm06, zmm22, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm07, zmm23, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm08, zmm24, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm09, zmm25, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm10, zmm26, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm11, zmm27, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm12, zmm28, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm13, zmm29, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm14, zmm30, zmm00);
	zmm00 = _mm512_fmadd_ps(zmm15, zmm31, zmm00);

	_mm512_store_ps(res, zmm00);
#else
	// matrix
	zmm00 = _mm512_load_pd(mtx);
	zmm01 = _mm512_load_pd(mtx + VECTOR_SIZE);
	zmm02 = _mm512_load_pd(mtx + VECTOR_SIZE * 2);
	zmm03 = _mm512_load_pd(mtx + VECTOR_SIZE * 3);
	zmm04 = _mm512_load_pd(mtx + VECTOR_SIZE * 4);
	zmm05 = _mm512_load_pd(mtx + VECTOR_SIZE * 5);
	zmm06 = _mm512_load_pd(mtx + VECTOR_SIZE * 6);
	zmm07 = _mm512_load_pd(mtx + VECTOR_SIZE * 7);
	// vector
	zmm08 = _mm512_extload_pd(vec, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm09 = _mm512_extload_pd(vec + 1, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm10 = _mm512_extload_pd(vec + 2, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm11 = _mm512_extload_pd(vec + 3, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm12 = _mm512_extload_pd(vec + 4, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm13 = _mm512_extload_pd(vec + 5, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm14 = _mm512_extload_pd(vec + 6, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm15 = _mm512_extload_pd(vec + 7, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);

	zmm16 = _mm512_mul_pd(zmm00, zmm08);
	zmm16 = _mm512_fmadd_pd(zmm01, zmm09, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm02, zmm10, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm03, zmm11, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm04, zmm12, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm05, zmm13, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm06, zmm14, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm07, zmm15, zmm16);

	_mm512_store_pd(res, zmm16);

#endif
}


void MVmSimdIntrinsic2(REAL* mtx, REAL* vec, REAL* res)
{
#ifdef SINGLE
	// matrix
	zmm00 = _mm512_load_ps(mtx);
	zmm01 = _mm512_load_ps(mtx + VECTOR_SIZE);
	zmm02 = _mm512_load_ps(mtx + VECTOR_SIZE * 2);
	zmm03 = _mm512_load_ps(mtx + VECTOR_SIZE * 3);
	zmm04 = _mm512_load_ps(mtx + VECTOR_SIZE * 4);
	zmm05 = _mm512_load_ps(mtx + VECTOR_SIZE * 5);
	zmm06 = _mm512_load_ps(mtx + VECTOR_SIZE * 6);
	zmm07 = _mm512_load_ps(mtx + VECTOR_SIZE * 7);
	// vector
	zmm08 = _mm512_extload_ps(vec, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm09 = _mm512_extload_ps(vec + 1, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm10 = _mm512_extload_ps(vec + 2, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm11 = _mm512_extload_ps(vec + 3, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm12 = _mm512_extload_ps(vec + 4, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm13 = _mm512_extload_ps(vec + 5, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm14 = _mm512_extload_ps(vec + 6, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm15 = _mm512_extload_ps(vec + 7, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);

	zmm16 = _mm512_mul_ps(zmm00, zmm08);
	zmm16 = _mm512_fmadd_ps(zmm01, zmm09, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm02, zmm10, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm03, zmm11, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm04, zmm12, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm05, zmm13, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm06, zmm14, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm07, zmm15, zmm16);

	// matrix
	zmm00 = _mm512_load_ps(mtx + VECTOR_SIZE * 8);
	zmm01 = _mm512_load_ps(mtx + VECTOR_SIZE * 9);
	zmm02 = _mm512_load_ps(mtx + VECTOR_SIZE * 10);
	zmm03 = _mm512_load_ps(mtx + VECTOR_SIZE * 11);
	zmm04 = _mm512_load_ps(mtx + VECTOR_SIZE * 12);
	zmm05 = _mm512_load_ps(mtx + VECTOR_SIZE * 13);
	zmm06 = _mm512_load_ps(mtx + VECTOR_SIZE * 14);
	zmm07 = _mm512_load_ps(mtx + VECTOR_SIZE * 15);
	// vector
	zmm08 = _mm512_extload_ps(vec + 8, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm09 = _mm512_extload_ps(vec + 9, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm10 = _mm512_extload_ps(vec + 10, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm11 = _mm512_extload_ps(vec + 11, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm12 = _mm512_extload_ps(vec + 12, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm13 = _mm512_extload_ps(vec + 13, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm14 = _mm512_extload_ps(vec + 14, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
	zmm15 = _mm512_extload_ps(vec + 15, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);

	zmm16 = _mm512_fmadd_ps(zmm00, zmm08, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm01, zmm09, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm02, zmm10, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm03, zmm11, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm04, zmm12, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm05, zmm13, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm06, zmm14, zmm16);
	zmm16 = _mm512_fmadd_ps(zmm07, zmm15, zmm16);

	_mm512_store_ps(res, zmm16);
#else
	// matrix
	zmm00 = _mm512_load_pd(mtx);
	zmm01 = _mm512_load_pd(mtx + VECTOR_SIZE);
	zmm02 = _mm512_load_pd(mtx + VECTOR_SIZE * 2);
	zmm03 = _mm512_load_pd(mtx + VECTOR_SIZE * 3);
	zmm04 = _mm512_load_pd(mtx + VECTOR_SIZE * 4);
	zmm05 = _mm512_load_pd(mtx + VECTOR_SIZE * 5);
	zmm06 = _mm512_load_pd(mtx + VECTOR_SIZE * 6);
	zmm07 = _mm512_load_pd(mtx + VECTOR_SIZE * 7);
	// vector
	zmm08 = _mm512_extload_pd(vec, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm09 = _mm512_extload_pd(vec + 1, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm10 = _mm512_extload_pd(vec + 2, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm11 = _mm512_extload_pd(vec + 3, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm12 = _mm512_extload_pd(vec + 4, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm13 = _mm512_extload_pd(vec + 5, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm14 = _mm512_extload_pd(vec + 6, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
	zmm15 = _mm512_extload_pd(vec + 7, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);

	zmm16 = _mm512_mul_pd(zmm00, zmm08);
	zmm16 = _mm512_fmadd_pd(zmm01, zmm09, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm02, zmm10, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm03, zmm11, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm04, zmm12, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm05, zmm13, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm06, zmm14, zmm16);
	zmm16 = _mm512_fmadd_pd(zmm07, zmm15, zmm16);

	_mm512_store_pd(res, zmm16);

#endif
}


double NextRandom()
{
	double number;
	do
	{
		number = (((double)std::rand()) / RAND_MAX - 0.5);
	} 
        while (std::fabs(number) < 1e-8);
	return number * 1e6;
}

void RandomFill(REAL* mtx, REAL* vec)
{
	std::srand(1986); //seed for random generator
	for (int k = 0; k < DIM; k++)
	{
		for (int i = 0; i < VECTOR_SIZE; i++)
		{
			vec[i + k*VECTOR_SIZE] = NextRandom();
			for (int j = 0; j < VECTOR_SIZE; j++)
			{
				mtx[i*VECTOR_SIZE + j + k*MATRIX_SIZE] = NextRandom();
			}
		}
	}
}

void OutputMtx(REAL* mtx, int nrows, int ncols)
{
	for (int i = 0; i < nrows; i++)
	{
	std::cout.precision(2);
	std::cout.width(11);

		for (int j = 0; j < ncols; j++)
			std::cout << mtx[i*ncols+j] << " ";
		std::cout << std::endl;
	}	
	std::cout << std::endl;
}

int main()
{
	REAL* matrix = (REAL*)_mm_malloc(sizeof(REAL)*VECTOR_SIZE*VECTOR_SIZE*DIM, ALIGNMENT);
	REAL* vector = (REAL*)_mm_malloc(sizeof(REAL)*VECTOR_SIZE*DIM, ALIGNMENT);
	REAL* vectorResult = (REAL*)_mm_malloc(sizeof(REAL)*VECTOR_SIZE*DIM, ALIGNMENT);

	std::cout << "Random Fill" << std::endl;
	RandomFill(matrix, vector);

	std::cout << "Warmup" << std::endl;
	for (int i = 0; i < 10; i++)
	{
		MVmDirect(matrix + i*MATRIX_SIZE, vector + i*VECTOR_SIZE, vectorResult + i*VECTOR_SIZE);
	}

	RandomFill(matrix, vector);

	double duration;

	duration = MeasuredRun(matrix,vector,vectorResult,MVmDirect);
	std::cout << "Direct: " << duration << std::endl;

	OutputMtx(vectorResult,VECTOR_SIZE/4,4);

	duration = MeasuredRun(matrix,vector,vectorResult,MVmDirectInv);
	std::cout << "Direct Invert Cycles: " << duration << std::endl;

	OutputMtx(vectorResult,VECTOR_SIZE/4,4);


	duration = MeasuredRun(matrix,vector,vectorResult,MVmSimdIntrinsic);
	std::cout << "SIMD: " << duration << std::endl;

	OutputMtx(vectorResult,VECTOR_SIZE/4,4);

	_mm_free(matrix);
	_mm_free(vector);
	_mm_free(vectorResult);

	return 0;
}
