#include <iostream>
#include <cstdlib>
#include <ctime>
#include <cstring>
#include <immintrin.h>
#include <cmath>

#ifdef SINGLE
#define VECTOR_SIZE 16
#else
#define VECTOR_SIZE 8
#endif

#define MATRIX_SIZE (VECTOR_SIZE*VECTOR_SIZE)

#ifdef SINGLE
typedef float REAL;
typedef __m512 type512;
#else
typedef double REAL;
typedef __m512d type512;
#endif

#define DIMM 100000*3

#ifdef SINGLE
#define DIM DIMM
#else
#define DIM DIMM*4
#endif

#define ALIGNMENT 64

REAL* matrixArray;
REAL* vectorArray;
REAL* vectorResultArray;

type512 zmm00;
type512 zmm01;
type512 zmm02;
type512 zmm03;
type512 zmm04;
type512 zmm05;
type512 zmm06;
type512 zmm07;
type512 zmm08;
type512 zmm09;
type512 zmm10;
type512 zmm11;
type512 zmm12;
type512 zmm13;
type512 zmm14;
type512 zmm15;
type512 zmm16;
type512 zmm17;
type512 zmm18;
type512 zmm19;
type512 zmm20;
type512 zmm21;
type512 zmm22;
type512 zmm23;
type512 zmm24;
type512 zmm25;
type512 zmm26;
type512 zmm27;
type512 zmm28;
type512 zmm29;
type512 zmm30;
type512 zmm31;


typedef void (* FUNC)(REAL *, REAL *, REAL *);

float MeasuredRun(REAL* imtx, REAL* ivec, REAL* ires, FUNC func)
{
	clock_t t = clock();
	for (int i = 0; i < DIM; i++)
	{
		func(imtx + i*MATRIX_SIZE, ivec + i*VECTOR_SIZE, ires + i*VECTOR_SIZE);
	}
	t = clock()-t;
	return ((float)t)/CLOCKS_PER_SEC;
}


void MVmDirect(REAL* imtx, REAL* ivec, REAL* ires)
{
	REAL* mtx, *vec, *res;
	int mtxOffset, vecOffset;
#ifdef SINGLE	
	const int sz = 4;
#else
	const int sz = 2;
#endif
	for (int l = 0; l < sz; l++)
	{
		for (int m = 0; m < sz; m++)
		{
			mtxOffset = l * 4 * VECTOR_SIZE+ m * 4;
			vecOffset = l * sz * 4 + m * 4;
			mtx = imtx + mtxOffset;
			vec = ivec + vecOffset;
			res = ires + vecOffset;
			for (int i = 0; i < 4; i++)
			{
				res[i] = 0.;
				for (int j = 0; j < 4; j++)
				{
					res[i] += mtx[j*VECTOR_SIZE + i] * vec[j];
				}
			}
		}
	}
}


void MVmSimdIntrinsic(REAL* mtx, REAL* invec, REAL* res)
{
#ifdef SINGLE
	const int N = 4;
#pragma novector
	for (int i = 0; i < N; i++)
	{
		REAL* vec = invec + i * VECTOR_SIZE;
		zmm00 = _mm512_load_ps(mtx + VECTOR_SIZE * i * N);
		zmm01 = _mm512_load_ps(mtx + VECTOR_SIZE * (i * N + 1));
		zmm02 = _mm512_load_ps(mtx + VECTOR_SIZE * (i * N + 2));
		zmm03 = _mm512_load_ps(mtx + VECTOR_SIZE * (i * N + 3));

		zmm08 = _mm512_load_ps(vec);
		zmm04 = _mm512_swizzle_ps(zmm08, _MM_SWIZ_REG_AAAA);
		zmm05 = _mm512_swizzle_ps(zmm08, _MM_SWIZ_REG_BBBB);
		zmm06 = _mm512_swizzle_ps(zmm08, _MM_SWIZ_REG_CCCC);
		zmm07 = _mm512_swizzle_ps(zmm08, _MM_SWIZ_REG_DDDD);

		zmm16 = _mm512_mul_ps(zmm00, zmm04);
		zmm16 = _mm512_fmadd_ps(zmm01, zmm05, zmm16);
		zmm16 = _mm512_fmadd_ps(zmm02, zmm06, zmm16);
		zmm16 = _mm512_fmadd_ps(zmm03, zmm07, zmm16);

		_mm512_store_ps(res+i*16, zmm16);
	}
#else
	const int N = 2;
#pragma novector
	for (int i = 0; i < N; i++)
	{
		REAL* vec = invec + i * VECTOR_SIZE;
		zmm00 = _mm512_load_pd(mtx + VECTOR_SIZE * i * N);
		zmm01 = _mm512_load_pd(mtx + VECTOR_SIZE * (i * N + 1));
		zmm02 = _mm512_load_pd(mtx + VECTOR_SIZE * (i * N + 2));
		zmm03 = _mm512_load_pd(mtx + VECTOR_SIZE * (i * N + 3));

		zmm08 = _mm512_load_pd(vec);
		zmm04 = _mm512_swizzle_pd(zmm08, _MM_SWIZ_REG_AAAA);
		zmm05 = _mm512_swizzle_pd(zmm08, _MM_SWIZ_REG_BBBB);
		zmm06 = _mm512_swizzle_pd(zmm08, _MM_SWIZ_REG_CCCC);
		zmm07 = _mm512_swizzle_pd(zmm08, _MM_SWIZ_REG_DDDD);

		zmm16 = _mm512_mul_pd(zmm00, zmm04);
		zmm16 = _mm512_fmadd_pd(zmm01, zmm05, zmm16);
		zmm16 = _mm512_fmadd_pd(zmm02, zmm06, zmm16);
		zmm16 = _mm512_fmadd_pd(zmm03, zmm07, zmm16);

		_mm512_store_pd(res+i*16, zmm16);
	}
#endif
}

void MVmSimdIntrinsicBlend(REAL* mtx, REAL* invec, REAL* res)
{
#ifdef SINGLE
	const int N = 4;
#pragma novector
	for (int i = 0; i < N; i++)
	{
		REAL* vec = invec + i * VECTOR_SIZE;
		zmm00 = _mm512_load_ps(mtx + VECTOR_SIZE * i * N);
		zmm01 = _mm512_load_ps(mtx + VECTOR_SIZE * (i * N + 1));
		zmm02 = _mm512_load_ps(mtx + VECTOR_SIZE * (i * N + 2));
		zmm03 = _mm512_load_ps(mtx + VECTOR_SIZE * (i * N + 3));

		zmm08 = _mm512_extload_ps(vec + 0, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm09 = _mm512_extload_ps(vec + 4, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm10 = _mm512_extload_ps(vec + 8, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm11 = _mm512_extload_ps(vec + 12, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm12 = _mm512_extload_ps(vec + 1, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm13 = _mm512_extload_ps(vec + 5, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm14 = _mm512_extload_ps(vec + 9, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm15 = _mm512_extload_ps(vec + 13, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);

		int mask32_0 = 0x00F0;
		int mask32_1 = 0xF000;
		int mask32_2 = 0xFF00;

		zmm06 = _mm512_mask_blend_ps(mask32_0, zmm08, zmm09);
		zmm07 = _mm512_mask_blend_ps(mask32_1, zmm10, zmm11);
		zmm04 = _mm512_mask_blend_ps(mask32_2, zmm06, zmm07);

		zmm06 = _mm512_mask_blend_ps(mask32_0, zmm12, zmm13);
		zmm07 = _mm512_mask_blend_ps(mask32_1, zmm14, zmm15);
		zmm05 = _mm512_mask_blend_ps(mask32_2, zmm06, zmm07);

		// 2x4 vectors
		zmm08 = _mm512_extload_ps(vec + 2, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm09 = _mm512_extload_ps(vec + 6, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm10 = _mm512_extload_ps(vec + 10, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm11 = _mm512_extload_ps(vec + 14, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm12 = _mm512_extload_ps(vec + 3, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm13 = _mm512_extload_ps(vec + 7, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm14 = _mm512_extload_ps(vec + 11, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);
		zmm15 = _mm512_extload_ps(vec + 15, _MM_UPCONV_PS_NONE, _MM_BROADCAST_1X16, _MM_HINT_NONE);

		zmm06 = _mm512_mask_blend_ps(mask32_0, zmm08, zmm09);
		zmm07 = _mm512_mask_blend_ps(mask32_1, zmm10, zmm11);
		zmm06 = _mm512_mask_blend_ps(mask32_2, zmm06, zmm07);

		zmm08 = _mm512_mask_blend_ps(mask32_0, zmm12, zmm13);
		zmm09 = _mm512_mask_blend_ps(mask32_1, zmm14, zmm15);
		zmm07 = _mm512_mask_blend_ps(mask32_2, zmm08, zmm09);

		zmm16 = _mm512_mul_ps(zmm00, zmm04);
		zmm16 = _mm512_fmadd_ps(zmm01, zmm05, zmm16);
		zmm16 = _mm512_fmadd_ps(zmm02, zmm06, zmm16);
		zmm16 = _mm512_fmadd_ps(zmm03, zmm07, zmm16);

		_mm512_store_ps(res+i*16, zmm16);
	}
#else
	const int N = 2;
#pragma novector
	for (int i = 0; i < N; i++)
	{
		REAL* vec = invec + i * VECTOR_SIZE;
		zmm00 = _mm512_load_pd(mtx + VECTOR_SIZE * i * N);
		zmm01 = _mm512_load_pd(mtx + VECTOR_SIZE * (i * N + 1));
		zmm02 = _mm512_load_pd(mtx + VECTOR_SIZE * (i * N + 2));
		zmm03 = _mm512_load_pd(mtx + VECTOR_SIZE * (i * N + 3));

		zmm08 = _mm512_extload_pd(vec + 0, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm09 = _mm512_extload_pd(vec + 4, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm10 = _mm512_extload_pd(vec + 1, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm11 = _mm512_extload_pd(vec + 5, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm12 = _mm512_extload_pd(vec + 2, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm13 = _mm512_extload_pd(vec + 6, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm14 = _mm512_extload_pd(vec + 3, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);
		zmm15 = _mm512_extload_pd(vec + 7, _MM_UPCONV_PD_NONE, _MM_BROADCAST_1X8, _MM_HINT_NONE);

		int mask = 0xF0;

		zmm16 = _mm512_mask_blend_pd(mask, zmm08, zmm09);
		zmm17 = _mm512_mask_blend_pd(mask, zmm10, zmm11);
		zmm18 = _mm512_mask_blend_pd(mask, zmm12, zmm13);
		zmm19 = _mm512_mask_blend_pd(mask, zmm14, zmm15);


		zmm20 = _mm512_mul_pd(zmm00, zmm16);
		zmm20 = _mm512_fmadd_pd(zmm01, zmm17, zmm20);
		zmm20 = _mm512_fmadd_pd(zmm02, zmm18, zmm20);
		zmm20 = _mm512_fmadd_pd(zmm03, zmm19, zmm20);

		_mm512_store_pd(res+i*8, zmm20);
	}
#endif
}


double NextRandom()
{
	double number;
	do
	{
		number = (((double)std::rand()) / RAND_MAX - 0.5);
	} 
        while (std::fabs(number) < 1e-8);
	return number * 1e6;
}

void RandomFill(REAL* mtx, REAL* vec)
{
	std::srand(1986); //seed for random generator
	for (int k = 0; k < DIM; k++)
	{
		for (int i = 0; i < VECTOR_SIZE; i++)
		{
			for(int ii = 0; ii < 4; ii++)
			{	
				vec[ii+i*4 + k*VECTOR_SIZE*4] = NextRandom();
			}	
			for (int j = 0; j < VECTOR_SIZE; j++)
			{
				mtx[i*VECTOR_SIZE + j + k*MATRIX_SIZE] = NextRandom();
			}
		}
	}
}


void OutputMtx(REAL* mtx, int nrows, int ncols)
{
	for (int i = 0; i < nrows; i++)
	{
	std::cout.precision(2);
	std::cout.width(11);

		for (int j = 0; j < ncols; j++)
			std::cout << mtx[i*ncols+j] << " ";
		std::cout << std::endl;
	}	
	std::cout << std::endl;
}

int main()
{
std::cout << "Start\n" ;
	REAL* matrix =	     (REAL*)_mm_malloc(sizeof(REAL)*MATRIX_SIZE*DIM, ALIGNMENT);
	REAL* vector = 	     (REAL*)_mm_malloc(sizeof(REAL)*VECTOR_SIZE*DIM*4, ALIGNMENT);
	REAL* vectorResult = (REAL*)_mm_malloc(sizeof(REAL)*VECTOR_SIZE*DIM*4, ALIGNMENT);
std::cout << "Random fill\n";
	RandomFill(matrix, vector);

std::cout << "Warmup\n";
	//warmup code
	for (int i = 0; i < 10; i++)
	{
		MVmDirect(matrix + i*MATRIX_SIZE, vector + i*VECTOR_SIZE*4, vectorResult + i*VECTOR_SIZE*4);
	}

	RandomFill(matrix, vector);

	double duration = MeasuredRun(matrix,vector,vectorResult,MVmDirect);
	std::cout << "Direct: " << duration << std::endl;

	OutputMtx(vectorResult, 4, 4);

	duration = MeasuredRun(matrix,vector,vectorResult,MVmSimdIntrinsicBlend);
	std::cout << "SIMD Blend: " << duration << std::endl;

	OutputMtx(vectorResult, 4, 4);


	duration = MeasuredRun(matrix,vector,vectorResult,MVmSimdIntrinsic);
	std::cout << "SIMD Swizzle: " << duration << std::endl;

	OutputMtx(vectorResult, 4, 4);


	_mm_free(matrix);
	_mm_free(vector);
	_mm_free(vectorResult);

	return 0;
}
